"""
Script for the various tower classes.
"""
import logging

import numpy as np
from ophyd import Component as Cmp
from ophyd.status import wait as status_wait

from .aerotech import (AeroBase, InterLinearAero, InterRotationAero,
                       LinearAero, RotationAero)
from .attocube import DiodeEcc, EccBase, GoniometerEcc, TranslationEcc
from .bragg import bragg_angle, bragg_energy
from .snddevice import SndDevice

logger = logging.getLogger(__name__)


class TowerBase(SndDevice):
    """
    Base tower class.
    """
    tab_whitelist = ['check_status', 'clear', 'disable', 'enable', 'energy',
                     'insert', 'inserted', 'position', 'remove', 'set_energy',
                     'status', 'stop', 'theta']

    def __init__(self, prefix, name=None, pos_inserted=None, pos_removed=None,
                 *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.pos_inserted = pos_inserted
        self.pos_removed = pos_removed
        self.desc_short = "".join([s[0] for s in self.desc.split(" ")])

        # Add Tower short name to desc
        for sig_name in self.component_names:
            signal = getattr(self, sig_name)
            if hasattr(signal, "desc"):
                signal.desc = "{0} {1}".format(self.desc_short, signal.desc)

    def set_energy(self, E, *args, **kwargs):
        """
        Placeholder for the energy setter. Implement for each TowerBase
        subclass.
        """
        pass

    @property
    def energy(self):
        """
        Returns the energy of the tower according to the angle of the
        arm.

        Returns
        -------
        E : float
            Energy of the delay line.
        """
        # Please forgive me, wasnt having a good day
        return int(np.round(bragg_energy(self.theta)*100))/100

    @energy.setter
    def energy(self, E):
        """
        Sets the theta of the tower to the desired energy. Alias for
        set_energy(E).

        Parameters
        ----------
        E : float
            Desired energy to set the tower to.
        """
        _ = self.set_energy(E)

    @property
    def position(self):
        """
        Current position of the tower. Implment this for each TowerBase
        subclass.
        """
        return None

    @property
    def theta(self):
        """
        Bragg angle the tower is currently set to maximize.

        Returns
        -------
        position : float
            Current position of the tower.
        """
        return self.position

    def insert(self, *args, **kwargs):
        """
        Moves the tower x motor to `self.pos_inserted`.

        Returns
        -------
        status : MoveStatus
            Status of the move.

        Raises
        ------
        ValueError
            If pos_inserted is set to None and insert() is called.
        """
        if self.pos_inserted is None:
            raise ValueError("Must set pos_inserted to use insert method.")
        return self.x.move(self.pos_inserted, *args, **kwargs)

    def remove(self, *args, **kwargs):
        """
        Moves the tower x motor to `self.pos_removed`.

        Returns
        -------
        status : MoveStatus
            Status of the move.

        Raises
        ------
        ValueError
            If pos_removed is set to None and remove() is called.
        """
        if self.pos_removed is None:
            raise ValueError("Must set pos_removed to use remove method.")
        return self.x.move(self.pos_removed, *args, **kwargs)

    @property
    def inserted(self):
        """
        Returns whether the tower is in the inserted position (or close to it).

        Returns
        -------
        inserted : bool
            Whether the tower is inserted or not.

        Raises
        ------
        ValueError
            If pos_inserted is set to None and inserted is called.
        """
        if self.pos_inserted is None:
            raise ValueError("Must set pos_inserted to check if inserted.")
        return np.isclose(self.pos_inserted, self.position, atol=0.1)

    def _get_move_positions(self, E):
        """
        Returns a list of positions that the energy motors should move to based
        on the inputted theta. Base implementation just returns a list of theta
        with length len(self._energy_motors).

        Parameters
        ----------
        E : float
            Energy to compute the motor move positions for.

        Returns
        -------
        positions : list
            List of positions each of the energy motors need to move to.
        """
        return [bragg_angle(E)] * len(self._energy_motors)

    def check_status(self, energy=None, length=None, no_raise=False):
        """
        Checks to make sure that all the energy motors are not in a bad state.
        Will include the delay motor if the delay argument is True.

        Parameters
        ----------
        energy : float or None, optional
            Energy to set the tower to.

        length : float or None, optional
            Length to set the delay stage to. (Doesnt apply for cc towers)

        no_raise : bool, optional
            Do not re-raise the attribute error for delay parameters.
        """
        # Create the list of motors and positions we will iterate through
        motors = []
        positions = []

        # Get all the energy parameters
        if energy is not None:
            motors += self._energy_motors
            _ = bragg_angle(energy)
            positions += self._get_move_positions(energy)

        # Get the delay parameters
        try:
            if length is not None:
                motors += [self.L]
                positions += length
        except AttributeError:
            if not no_raise:
                raise

        # Check that we can move all the motors
        for motor, position in zip(motors, positions):
            try:
                motor.check_status(position)
            except Exception as e:
                err = "Motor {0} got an exception: {1}".format(motor.desc, e)
                logger.error(err)
                raise e

    def stop(self):
        """
        Stops the motions of all the motors.
        """
        self._apply_all("stop", (AeroBase, EccBase), print_set=False)

    def enable(self):
        """
        Enables all the aerotech motors.
        """
        self._apply_all("enable", (AeroBase, EccBase), print_set=False)

    def disable(self):
        """
        Disables all the aerotech motors.
        """
        self._apply_all("disable", (AeroBase, EccBase), print_set=False)

    def clear(self):
        """
        Disables all the aerotech motors.
        """
        self._apply_all("clear", AeroBase, print_set=False)

    def status(self, status="", offset=0, print_status=True, newline=False,
               short=True):
        """
        Returns the status of the tower.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.

        newline : bool, optional
            Adds a new line to the end of the string.

        Returns
        -------
        status : str
            Status string.
        """
        if short:
            # Header
            status += "\n{0}{1}\n{2}{3}".format(
                " "*offset, self.desc, " "*(offset+2), "-"*50)

            # Aerotech body
            status_list_aero = self._apply_all(
                "status", AeroBase, offset=offset+2, print_status=False,
                short=True)
            if status_list_aero:
                # Aerotech header
                status += (
                    "\n{0}{1:<16}|{2:^16}|{3:^16}\n{4}{5}".format(
                        " "*(offset+2), "Motor", "Position", "Dial", " "*(offset+2), "-"*50
                    )
                )
                status += "".join(status_list_aero)

            # Attocube body
            status_list_atto = self._apply_all(
                "status", EccBase, offset=offset+2, print_status=False,
                short=True)
            if status_list_atto:
                # Attocube Header
                status += (
                    "\n{0}{1}\n{2}{3:<16}|{4:^16}|{5:^16}\n{6}{7}".format(
                        " "*(offset+2), "-"*50, " " * (offset+2),
                        "Motor", "Position", "Reference", " "*(offset+2), "-"*50)
                )
                status += "".join(status_list_atto)

        else:
            status += "{0}{1}:\n{2}{3}\n".format(
                " "*offset, self.desc, " "*offset, "-"*(len(self.desc)+1)
            )
            status_list = self._apply_all("status", (AeroBase, EccBase),
                                          offset=offset+2, print_status=False)
            status += "".join(status_list)

        if newline:
            status += "\n"
        if print_status:
            logger.info(status)
        else:
            return status


class DelayTower(TowerBase):
    """
    Delay Tower

    Components

    tth : RotationAeroInterlocked
        Rotation axis of the entire delay arm.

    th1 : RotationAero
        Rotation axis of the static crystal.

    th2 : RotationAero
        Rotation axis of the delay crystal.

    x : LinearAero
        Linear stage for insertion/bypass of the tower.

    L : LinearAero
        Linear stage for the delay crystal.

    y1 : TranslationEcc
        Y translation for the static crystal.

    y2 : TranslationEcc
        Y translation for the delay crystal.

    chi1 : GoniometerEcc
        Goniometer on static crystal.

    chi2 : GoniometerEcc
        Goniometer on delay crystal.

    dh : DiodeEcc
        Diode insertion motor.

    diode : HamamatsuDiode
        Diode between the static and delay crystals.

    temp : OmegaRTD
        RTD temperature sensor for the nitrogen.
    """
    tab_whitelist = ['length', 'position', 'set_energy', 'set_length', 'theta']

    # Rotation stages
    tth = Cmp(InterRotationAero, ":TTH", desc="TTH")
    th1 = Cmp(RotationAero, ":TH1", desc="TH1")
    th2 = Cmp(RotationAero, ":TH2", desc="TH2")

    # Linear stages
    x = Cmp(InterLinearAero, ":X", desc="X")
    L = Cmp(InterLinearAero, ":L", desc="L")

    # # Y Crystal motion
    y1 = Cmp(TranslationEcc, ":Y1", desc="Y1")
    y2 = Cmp(TranslationEcc, ":Y2", desc="Y2")

    # Chi motion
    chi1 = Cmp(GoniometerEcc, ":CHI1", desc="CHI1")
    chi2 = Cmp(GoniometerEcc, ":CHI2", desc="CHI2")

    # Diode motion
    dh = Cmp(DiodeEcc, ":DH", desc="DH")

    # # Diode
    # diode = Cmp(HamamatsuDiode, ":DIODE", desc="Tower Diode")

    # # Temperature monitor
    # temp = Cmp(OmegaRTD, ":TEMP", desc="Tower RTD")

    def __init__(self, prefix, *args, **kwargs):
        super().__init__(prefix, *args, **kwargs)
        self._energy_motors = [self.tth, self.th1, self.th2]

    @property
    def position(self):
        """
        Returns the theta position of the arm (tth) in degrees.

        Returns
        -------
        position : float
            Position of the arm in degrees.
        """
        return self.tth.position

    def _get_move_positions(self, E):
        """
        Returns the list of positions that each of the energy motors need to
        move to based on the inputted theta. tth moves to 2*theta while all
        the other motors move to theta.

        Parameters
        ----------
        E : float
            Energy to compute the motor move positions for.

        Returns
        -------
        positions : list
            List of positions each of the energy motors need to move to.
        """
        # Convert to theta
        theta = bragg_angle(E=E)

        # Get the positions
        positions = []
        for motor in self._energy_motors:
            if motor is self.tth:
                positions.append(2*theta)
            else:
                positions.append(theta)
        return positions

    def set_energy(self, E, wait=False, check_status=True):
        """
        Sets the angles of the crystals in the delay line to maximize the
        inputted energy.

        Parameters
        ---------
        E : float
            Energy to use for the system.

        wait : bool, optional
            Wait for each motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.
        """
        # Check to make sure the motors are in a valid state to move
        if check_status:
            self.check_status(energy=E)

        # Perform the move
        status = [motor.move(pos, wait=False, check_status=False) for
                  motor, pos in zip(self._energy_motors,
                                    self._get_move_positions(E))]

        # Wait for the motions to finish
        if wait:
            for s in status:
                logger.info("Waiting for {} to finish move ...".format(
                    s.device.name))
                status_wait(s)

        return status

    def set_length(self, position, wait=False, *args, **kwargs):
        """
        Sets the position of the linear delay stage in mm.

        Parameters
        ----------
        position : float
            Position to move the delay motor to.

        wait : bool, optional
            Wait for motion to complete before returning the console.

        Returns
        -------
        status : MoveStatus
            Status object of the move.
        """
        return self.L.move(position, wait=wait, *args, **kwargs)

    @property
    def length(self):
        """
        Returns the position of the linear delay stage (L) in mm.

        Returns
        -------
        position : float
            Position in mm of the linear delay stage.
        """
        return self.L.position

    @length.setter
    def length(self, position):
        """
        Sets the position of the linear delay stage in mm.

        Parameters
        ----------
        position : float
            Position to move the delay motor to.
        """
        _ = self.L.mv(position)

    @property
    def theta(self):
        """
        Bragg angle the tower is currently set to maximize.

        Returns
        -------
        position : float
            Current position of the tower.
        """
        return self.position/2


class ChannelCutTower(TowerBase):
    """
    Channel Cut tower.

    Components

    th : RotationAero
        Rotation stage of the channel cut crystal

    x : LinearAero
        Translation stage of the tower
    """
    tab_whitelist = ['position', 'set_energy']

    # Rotation
    th = Cmp(RotationAero, ":TH", desc="TH")

    # Translation
    x = Cmp(LinearAero, ":X", desc="X")

    def __init__(self, prefix, *args, **kwargs):
        super().__init__(prefix, *args, **kwargs)
        self._energy_motors = [self.th]

    @property
    def position(self):
        """
        Returns the theta position of the crystal (th) in degrees.

        Returns
        -------
        position : float
            Position of the arm in degrees.
        """
        return self.th.position

    def set_energy(self, E, wait=False, check_status=True):
        """
        Sets the angles of the crystals in the channel cut line to maximize the
        inputted energy.

        Parameters
        ---------
        E : float
            Energy to use for the system.

        wait : bool, optional
            Wait for motion to complete before returning the console.

        check_status : bool, optional
            Check if the motors are in a valid state to move.
        """
        # Convert to theta
        theta = bragg_angle(E=E)

        # Check to make sure the motors are in a valid state to move
        if check_status:
            self.check_status(E)

        # Perform the move
        status = self.th.move(theta, wait=wait)
        return status



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import time
from collections import OrderedDict

import pytest
from ophyd.device import Device

from hxrsnd import tower
from hxrsnd.exceptions import MotorDisabled, MotorFaulted
from hxrsnd.sndsystem import ChannelCutTower, DelayTower

from .conftest import fake_device, get_classes_in_module

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("dev", get_classes_in_module(tower, Device))
def test_devices_instantiate_and_run_ophyd_functions(dev):
    device = fake_device(dev, "TEST:SND:T1")
    assert(isinstance(device.read(), OrderedDict))
    assert(isinstance(device.read_configuration(), OrderedDict))


def test_DelayTower_does_not_move_if_motors_not_ready():
    tower = fake_device(DelayTower, "TEST:SND:T1")
    tower.disable()
    time.sleep(.5)
    tower.tth.user_setpoint.sim_set_limits((-100, 100))
    tower.th1.user_setpoint.sim_set_limits((-100, 100))
    tower.th2.user_setpoint.sim_set_limits((-100, 100))

    tower.tth.user_setpoint.check_value = lambda x: None
    tower.th1.user_setpoint.check_value = lambda x: None
    tower.th2.user_setpoint.check_value = lambda x: None

    with pytest.raises(MotorDisabled):
        tower.energy = 10
    tower.enable()
    tower.tth.axis_fault.sim_put(True)
    with pytest.raises(MotorFaulted):
        tower.energy = 10


def test_ChannelCutTower_does_not_move_if_motors_not_ready():
    tower = fake_device(ChannelCutTower, "TEST:SND:T1")
    tower.disable()
    time.sleep(.5)
    tower.th.user_setpoint.sim_set_limits((-100, 100))
    tower.th.user_setpoint.check_value = lambda x: None

    with pytest.raises(MotorDisabled):
        tower.energy = 10
    tower.enable()
    tower.th.axis_fault.sim_put(True)
    with pytest.raises(MotorFaulted):
        tower.energy = 10



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from collections import OrderedDict
from copy import deepcopy

import numpy as np
import pandas as pd
import pytest
from bluesky.preprocessors import run_wrapper
from ophyd.device import Device
from ophyd.sim import SynAxis

from hxrsnd import sndmotor

from ..exceptions import InputError
from ..plans.scans import centroid_scan
from ..sndmotor import CalibMotor
from .conftest import SynCamera, fake_device, get_classes_in_module

logger = logging.getLogger(__name__)
rtol = 1e-6                             # Numpy relative tolerance


@pytest.mark.parametrize("dev", get_classes_in_module(sndmotor, Device))
def test_sndmotor_devices_instantiate_and_run_ophyd_functions(dev):
    device = fake_device(dev)
    assert(isinstance(device.read(), OrderedDict))
    assert(isinstance(device.read_configuration(), OrderedDict))


def test_CalibMotor_configure_raises_errors_on_bad_inputs():
    dev = CalibMotor("TST", name="test")
    config = deepcopy(dev.read_configuration())

    # Define a quick function to test value equivalence
    def test_config_change(cfg1, cfg2):
        for key1, key2 in zip(cfg1.keys(), cfg2.keys()):
            if cfg1[key1]['value'] != cfg2[key2]['value']:
                return False
        return True

    # No errors if we dont pass any calib or motors
    assert dev.configure()
    assert test_config_change(config, dev.read_configuration())

    # TypeError when the calibration is not a dataframe
    with pytest.raises(TypeError):
        dev.configure(calib="test")
    # Assert the calibration wasnt changed
    assert test_config_change(config, dev.read_configuration())

    # InputError if we pass a dataframe but no motors
    with pytest.raises(InputError):
        dev.configure(calib=pd.DataFrame(columns=['a']))
    # Assert the calibration wasnt changed
    assert test_config_change(config, dev.read_configuration())

    # InputError if we pass a different number of motors as calib columns
    with pytest.raises(InputError):
        dev.configure(calib=pd.DataFrame(columns=['a', 'b']),
                      motors=[SynAxis(name="test")])
    # Assert the calibration wasnt changed
    assert test_config_change(config, dev.read_configuration())
    with pytest.raises(InputError):
        dev.configure(calib=pd.DataFrame(columns=['a']),
                      motors=[SynAxis(name="test_1"), SynAxis(name="test_2")])
    # Assert the calibration wasnt changed
    assert test_config_change(config, dev.read_configuration())


@pytest.mark.parametrize("scale", [None, [1]])
@pytest.mark.parametrize("start", [None, [1]])
@pytest.mark.parametrize("scan", [None, pd.DataFrame(columns=['b'])])
def test_CalibMotor_configure_works_with_good_inputs(scale, start, scan):
    dev = CalibMotor("TST", name="test")

    # There should be no errors with this configure
    assert dev.configure(calib=pd.DataFrame(columns=['a']),
                         motors=[SynAxis(name="test_1")],
                         scan=scan,
                         scale=scale,
                         start=start)

    config = dev.read_configuration()
    # Assert the extra values are in the cofiguration
    assert config['scan']['value'] is scan
    assert config['scale']['value'] == scale
    assert config['start']['value'] == start


@pytest.mark.parametrize("calib", [None,
                                   pd.DataFrame(columns=['a']),
                                   pd.DataFrame(columns=['a', 'b'])])
@pytest.mark.parametrize("motors", [None,
                                    [SynAxis(name='a')],
                                    [SynAxis(name='a'), SynAxis(name='b')]])
def test_CalibMotor_has_calib_correctly_indicates_there_is_a_valid_calibration(
        calib, motors):
    dev = CalibMotor("TST", name="test")
    # Force replace the underlying calibration to use the inputs
    dev._calib['calib']['value'] = calib
    dev._calib['motors']['value'] = motors

    # There must be both a calibration and motors, and they are the same length
    expected_logic = bool((calib is not None and motors) and
                          (len(calib.columns) == len(motors)))
    assert dev.has_calib == expected_logic


def test_CalibMotor_calibration_returns_correct_parameters():
    dev = CalibMotor("TST", name="test")
    # Should be none if there is no calibration
    assert dev.calibration is None

    # Create the parmeters we will pass
    calib = pd.DataFrame(columns=['a'])
    scan = pd.DataFrame(columns=['b'])
    motors = [SynAxis(name='test')]
    scale = [1]
    start = [0]

    # Configure and test the resulting calibration
    dev.configure(calib=calib, scan=scan, motors=motors, scale=scale,
                  start=start)
    calibration = dev.calibration
    assert calibration['calib'].equals(calib)
    assert calibration['scan'].equals(scan)
    assert calibration['motors'] == [m.name for m in motors]
    assert calibration['scale'] == scale
    assert calibration['start'] == start


def test_CalibMotor_calibrates_correctly(fresh_RE, get_calib_motor):
    # Define all the needed variables
    motor = get_calib_motor
    calib_motors = motor.calib_motors
    camera = SynCamera(*motor.calib_motors, motor, name="camera")
    centroids = [camera.centroid_x, camera.centroid_y]
    start, stop, steps = -1, 1, 5

    # Perform the calibration
    motor.calibrate(start, stop, steps, RE=fresh_RE, detector=camera,
                    average=1, tolerance=0, confirm_overwrite=False)

    # Grab the resulting calibration parameters
    df_calib = motor.calibration['calib']
    df_scan = motor.calibration['scan']

    # Expected positions of the centroids are the first positions
    expected_centroids = df_scan[[c.name for c in centroids]].iloc[0]

    for i in range(len(df_scan)):
        # Move to the scan position for the main motor
        motor.set(df_calib[motor.motor_fields[0]].iloc[i])

        for cmotor, exp_cent, cent in zip(calib_motors, expected_centroids,
                                          centroids):
            # Move to the abosolute corrected position
            cmotor.set(df_calib[cmotor.name+"_post"].iloc[i])
            # Check the centroids are where they should be
            assert np.isclose(cent.get(), exp_cent, rtol=rtol)


def test_CalibMotor_move_compensation(fresh_RE, get_calib_motor):
    # Define all the needed variables
    motor = get_calib_motor
    camera = SynCamera(*motor.calib_motors, motor, name="camera")
    centroids = [camera.centroid_x, camera.centroid_y]
    start, stop, steps = -1, 1, 5

    # Create the plan
    def test_plan():
        # Perform the calibration
        df_scan = (yield from centroid_scan(
            camera,
            motor,
            start, stop, steps,
            motor_fields=motor.motor_fields,
            detector_fields=motor.detector_fields))

        # Expected positions of the centroids are the first positions
        df_centroids = df_scan[[c.name for c in centroids]]
        assert (df_centroids == df_centroids.iloc[0]).all().all()

    # Calibrate the motors
    motor.calibrate(start, stop, steps, RE=fresh_RE, detector=camera,
                    average=1, tolerance=0, confirm_overwrite=False)
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))



"""
Aerotech devices
"""
import logging
import os

import numpy as np
from ophyd import Component as Cmp
from ophyd import FormattedComponent as FrmCmp
from ophyd.signal import EpicsSignal, EpicsSignalRO, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError

from .exceptions import (BadN2Pressure, MotorDisabled, MotorFaulted,
                         MotorStopped)
from .pneumatic import PressureSwitch
from .sndmotor import SndEpicsMotor
from .utils import absolute_submodule_path, as_list, stop_on_keyboardinterrupt

logger = logging.getLogger(__name__)


class AeroBase(SndEpicsMotor):
    """
    Base Aerotech motor class.

    Components

    power : EpicsSignal, ".CNEN"
        Enables or disables power to the axis.

    retries : EpicsSignalRO, ".RCNT"
        Number of retries attempted.

    retries_max : EpicsSignal, ".RTRY"
        Maximum number of retries.

    retries_deadband : EpicsSignal, ".RDBD"
        Tolerance of each retry.

    axis_fault : EpicsSignalRO, ":AXIS_FAULT"
        Fault readback for the motor.

    axis_status : EpicsSignalRO, ":AXIS_STATUS"
        Status readback for the motor.

    clear_error : EpicsSignal, ":CLEAR"
        Clear error signal.

    config : EpicsSignal, ":CONFIG"
        Signal to reconfig the motor.
    """
    # Remove when these have been figured out
    low_limit_switch = Cmp(Signal)
    high_limit_switch = Cmp(Signal)

    power = Cmp(EpicsSignal, ".CNEN")
    retries = Cmp(EpicsSignalRO, ".RCNT")
    retries_max = Cmp(EpicsSignal, ".RTRY")
    retries_deadband = Cmp(EpicsSignal, ".RDBD")
    axis_status = Cmp(EpicsSignalRO, ":AXIS_STATUS")
    axis_fault = Cmp(EpicsSignalRO, ":AXIS_FAULT")
    clear_error = Cmp(EpicsSignal, ":CLEAR")
    config = Cmp(EpicsSignal, ":CONFIG")
    zero_all_proc = Cmp(EpicsSignal, ":ZERO_P.PROC")
    home_forward = Cmp(EpicsSignal, ".HOMF")
    home_reverse = Cmp(EpicsSignal, ".HOMR")
    dial = Cmp(EpicsSignalRO, ".DRBV")
    state_component = Cmp(EpicsSignal, ".SPMG")

    def __init__(self, prefix, name=None, *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.motor_done_move.unsubscribe(self._move_changed)
        self.user_readback.unsubscribe(self._pos_changed)
        self.configuration_attrs.append("power")
        self._state_list = ["Stop", "Pause", "Move", "Go"]

    def _status_print(self, status, msg=None, ret_status=False, print_set=True,
                      timeout=None, wait=True, reraise=False):
        """
        Internal method that optionally returns the status object and optionally
        prints a message about the set. If a message is passed but print_set is
        False then the message is logged at the debug level.

        Parameters
        ----------
        status : StatusObject or list
            The inputted status object.

        msg : str or None, optional
            Message to print if print_set is True.

        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        wait : bool, optional
            Wait for the status to complete.

        reraise : bool, optional
            Raise the RuntimeError in the except.

        Returns
        -------
        Status
            Inputted status object.
        """
        try:
            # Wait for the status to complete
            if wait:
                for s in as_list(status):
                    status_wait(s, timeout)

            # Notify the user
            if msg is not None:
                if print_set:
                    logger.info(msg)
                else:
                    logger.debug(msg)
            if ret_status:
                return status

        # The operation failed for some reason
        except RuntimeError:
            error = "Operation completed, but reported an error."
            logger.error(error)
            if reraise:
                raise

    @stop_on_keyboardinterrupt
    def homf(self, ret_status=False, print_set=True, check_status=True):
        """
        Home the motor forward.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        Returns
        -------
        Status : StatusObject
            Status of the set.
        """
        # Check the motor status
        if check_status:
            self.check_status()
        status = self.home_forward.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Homing '{0}' forward.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    @stop_on_keyboardinterrupt
    def homr(self, ret_status=False, print_set=True, check_status=True):
        """
        Home the motor in reverse.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        Returns
        -------
        Status : StatusObject
            Status of the set.
        """
        # Check the motor status
        if check_status:
            self.check_status()
        status = self.home_reverse.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Homing '{0}' in reverse.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    def move(self, position, wait=False, check_status=True, timeout=None, *args,
             **kwargs):
        """
        Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to.

        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.

        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus
            Status object for the move.

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`

        ValueError
            On invalid positions

        RuntimeError
            If motion fails other than timing out
        """
        # Check the motor status
        if check_status:
            self.check_status(position)
        logger.debug("Moving {0} to {1}".format(self.name, position))
        return super().move(position, wait=wait, timeout=timeout, *args,
                            **kwargs)

    def mv(self, position, wait=True, print_move=True, *args, **kwargs):
        """
        Move to a specified position, optionally waiting for motion to
        complete. mv() is different from move() by catching all the common
        exceptions that this motor can raise and just raises a logger
        warning. Therefore if building higher level functionality, do not
        use this method and use move() instead otherwise none of these
        exceptions will propagate to it.

        Parameters
        ----------
        position
            Position to move to.

        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        print_move : bool, optional
            Print a short statement about the move.

        Exceptions Caught
        -----------------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.

        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised if the motor is stopped and the move is requested.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        try:
            status = super().mv(position, wait=wait, *args, **kwargs)

            # Notify the user that a motor has completed or the command is sent
            if print_move:
                if wait:
                    logger.info("Move completed for '{0}'.".format(self.desc))
                else:
                    logger.info(
                        "Move command sent to '{0}'.".format(self.desc))
            return status

        # Catch all the common motor exceptions
        except LimitError:
            logger.warning("Requested move '{0}' is outside the soft limits "
                           "{1} for motor {2}".format(position, self.limits,
                                                      self.desc))
        except MotorDisabled:
            logger.warning("Cannot move - motor {0} is currently disabled. Try "
                           "running 'motor.enable()'.".format(self.desc))
        except MotorFaulted:
            logger.warning("Cannot move - motor {0} is currently faulted. Try "
                           "running 'motor.clear()'.".format(self.desc))
        except MotorStopped:
            logger.warning("Cannot move - motor {0} is currently stopped. Try "
                           "running 'motor.state=\"Go\"'.".format(self.desc))

    def check_status(self, position=None):
        """
        Checks the status of the motor to make sure it is ready to move. Checks
        the current position of the motor, and if a position is provided it also
        checks that position.

        Parameters
        ----------
        position : float, optional
            Position to check for validity.

        Raises
        ------
        MotorDisabled
            If the motor is disabled.

        MotorFaulted
            If the motor is faulted.

        MotorStopped
            If the motor is stopped.
        """
        if not self.enabled:
            err = "Motor '{0}' is currently disabled".format(self.desc)
            logger.error(err)
            raise MotorDisabled(err)

        if self.faulted:
            err = "Motor '{0}' is currently faulted.".format(self.desc)
            logger.error(err)
            raise MotorFaulted(err)

        if self.state == "Stop":
            err = "Motor '{0}' is currently stopped.".format(self.desc)
            logger.error(err)
            raise MotorStopped(err)

        # Check if the current position is valid
        self.check_value(self.position)
        # Check if the move position is valid
        if position:
            self.check_value(position)

    def set_position(self, position_des, print_set=True):
        """
        Sets the current position to be the inputted position by changing the
        offset.

        Parameters
        ----------
        position_des : float
            The desired current position.
        """
        # Print to console or just to the log
        if print_set:
            log_level = logger.info
        else:
            log_level = logger.debug

        log_level("'{0}' previous position: {0}, offset: {1}".format(
            self.position, self.offset))
        self.offset += position_des - self.position
        log_level("'{0}' New position: {0}, offset: {1}".format(self.position,
                                                                self.offset))

    def enable(self, ret_status=False, print_set=True):
        """
        Enables the motor power.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the power signal.
        """
        status = self.power.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Enabled motor '{0}'.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    def disable(self, ret_status=False, print_set=True):
        """
        Disables the motor power.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the power signal.
        """
        status = self.power.set(0, timeout=self.set_timeout)
        return self._status_print(status, "Disabled motor '{0}'.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    @property
    def enabled(self):
        """
        Returns if the motor is curently enabled.

        Returns
        -------
        enabled : bool
            True if the motor is powered, False if not.
        """
        return bool(self.power.get())

    def clear(self, ret_status=False, print_set=True):
        """
        Clears the motor error.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the clear_error signal.
        """
        status = self.clear_error.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Cleared motor '{0}'.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    def reconfig(self, ret_status=False, print_set=True):
        """
        Re-configures the motor.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the config signal.
        """
        status = self.config.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Reconfigured motor '{0}'.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    @property
    def faulted(self):
        """
        Returns the current fault with the motor.

        Returns
        -------
        faulted
            Fault enumeration.
        """
        if self.axis_fault.connected:
            return bool(self.axis_fault.get())
        else:
            return None

    def zero_all(self, ret_status=False, print_set=True):
        """
        Sets the current position to be the zero position of the motor.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        status : StatusObject
            Status object for the set.
        """
        status = self.zero_all_proc.set(1, timeout=self.set_timeout)
        return self._status_print(status, "Zeroed motor '{0}'.".format(
            self.desc), print_set=print_set, ret_status=ret_status)

    @property
    def state(self):
        """
        Returns the state of the motor. State can be one of the following:
            'Stop', 'Pause', 'Move', 'Go'

        Returns
        -------
        state : str
            The current state of the motor
        """
        return self._state_list[self.state_component.get()]

    @state.setter
    def state(self, val, ret_status=False, print_set=True):
        """
        Sets the state of the motor. Inputted state can be one of the following
        states or the index of the desired state:
            'Stop', 'Pause', 'Move', 'Go'
        Alias for set_state((val, False, True)

        Parameters
        ----------
        val : int or str

        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the state signal.
        """
        try:
            return self.set_state(val, ret_status, print_set)
        except ValueError:
            logger.info("State must be one of the following: {0}".format(
                self._state_list))

    def set_state(self, state, ret_status=True, print_set=False):
        """
        Sets the state of the motor. Inputted state can be one of the following
        states or the index of the desired state:
            'Stop', 'Pause', 'Move', 'Go'

        Parameters
        ----------
        val : int or str

        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for setting the state signal.
        """
        # Make sure it is in title case if it's a string
        val = state
        if isinstance(state, str):
            val = state.title()

        # Make sure it is a valid state or enum
        if val not in self._state_list + list(range(len(self._state_list))):
            error = "Invalid state inputted: '{0}'.".format(val)
            logger.error(error)
            raise ValueError(error)

        # Lets enforce it's a string or value
        status = self.state_component.set(val, timeout=self.set_timeout)

        return self._status_print(
            status, "Changed state of '{0} to '{1}'.".format(self.desc, val),
            print_set=print_set, ret_status=ret_status)

    def ready_motor(self, ret_status=False, print_set=True):
        """
        Sets the motor to the ready state by clearing any errors, enabling it,
        and setting the state to be 'Go'.

        Parameters
        ----------
        ret_status : bool, optional
            Return the status object of the set.

        print_set : bool, optional
            Print a short statement about the set.

        Returns
        -------
        Status
            The status object for all the sets.
        """
        status = [self.clear(ret_status=True, print_set=False)]
        status.append(self.enable(ret_status=True, print_set=False))
        status.append(self.set_state("Go", ret_status=True, print_set=False))
        return self._status_print(
            status, "Motor '{0}' is now ready to move.".format(self.desc),
            print_set=print_set, ret_status=ret_status)

    def expert_screen(self, print_msg=True):
        """
        Launches the expert screen for the motor.

        Parameters
        ----------
        print_msg : bool, optional
            Prints that the screen is being launched.
        """
        path = absolute_submodule_path(
            "hxrsnd/screens/motor_expert_screens.sh")
        if print_msg:
            logger.info("Launching expert screen.")
        os.system("{0} {1} {2} &".format(path, self.prefix, "aerotech"))

    def status(self, status="", offset=0, print_status=True, newline=False,
               short=False):
        """
        Returns the status of the device.

        Parameters
        ----------
        status : str, optional
            The string to append the status to.

        offset : int, optional
            Amount to offset each line of the status.

        print_status : bool, optional
            Determines whether the string is printed or returned.

        newline : bool, optional
            Adds a new line to the end of the string.

        Returns
        -------
        status : str
            Status string.
        """
        if short:
            status += "\n{0}{1:<16}|{2:^16.3f}|{3:^16.3f}".format(
                " "*offset, self.desc, self.position, self.dial.get())
        else:
            status += "{0}{1}\n".format(" "*offset, self.desc)
            status += "{0}PV: {1:>25}\n".format(" "*(offset+2), self.prefix)
            status += "{0}Enabled: {1:>20}\n".format(" "*(offset+2),
                                                     str(self.enabled))
            status += "{0}Faulted: {1:>20}\n".format(" "*(offset+2),
                                                     str(self.faulted))
            status += "{0}State: {1:>22}\n".format(" "*(offset+2),
                                                   str(self.state))
            status += "{0}Position: {1:>19}\n".format(" "*(offset+2),
                                                      np.round(self.wm(), 6))
            status += "{0}Dial: {1:>23}\n".format(" "*(offset+2),
                                                  np.round(self.dial.get(),
                                                           6))
            status += "{0}Limits: {1:>21}\n".format(
                " "*(offset+2), str((int(self.low_limit),
                                     int(self.high_limit))))

        if newline:
            status += "\n"
        if print_status is True:
            logger.info(status)
        else:
            return status


class InterlockedAero(AeroBase):
    """
    Linear Aerotech stage that has the additional move check for the pressure
    status.
    """
    # To do the internel pressure check
    _pressure = FrmCmp(PressureSwitch, "{self._prefix}:N2:{self._tower}")

    def __init__(self, prefix, *args, **kwargs):
        self._tower = prefix.split(":")[-2]
        self._prefix = ":".join(prefix.split(":")[:2])
        super().__init__(prefix, *args, **kwargs)

    def check_status(self, *args, **kwargs):
        """
        Status check that also checks if the pressure measured by the pressure
        switch is good.

        Parameters
        ----------
        position : float
            Position to check for validity.

        Raises
        ------
        MotorDisabled
            If the motor is disabled.

        MotorFaulted
            If the motor is faulted.

        MotorStopped
            If the motor is stopped.

        BadN2Pressure
            If the pressure in the tower is bad.
        """
        if self._pressure.bad:
            err = "Cannot move - Pressure in {0} is bad.".format(self._tower)
            logger.error(err)
            raise BadN2Pressure(err)
        super().check_status(*args, **kwargs)

    def mv(self, position, *args, **kwargs):
        """
        Move to a specified position, optionally waiting for motion to
        complete. mv() is different from move() by catching all the common
        exceptions that this motor can raise and just raises a logger
        warning. Therefore if building higher level functionality, do not
        use this method and use move() instead otherwise none of these
        exceptions will propagate to it.

        Parameters
        ----------
        position
            Position to move to.

        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        ret_status : bool, optional
            Return the status object of the move.

        print_move : bool, optional
            Print a short statement about the move.

        Exceptions Caught
        -----------------
        LimitError
            Error raised when the inputted position is beyond the soft limits.

        MotorDisabled
            Error raised if the motor is disabled and move is requested.


        MotorFaulted
            Error raised if the motor is disabled and the move is requested.

        MotorStopped
            Error raised If the motor is stopped and a move is requested.

        BadN2Pressure
            Error raised if the pressure in the tower is bad.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        try:
            return self.move(position, *args, **kwargs)
        # Catch a bad pressure setting.
        except BadN2Pressure:
            logger.warning("Cannot move - pressure in tower {0} is bad.".format(
                self._tower))


class LinearAero(AeroBase):
    """
    Class for the aerotech linear stage.
    """
    pass


class InterLinearAero(InterlockedAero, LinearAero):
    """
    Class for the interlocked aerotech linear stage.
    """
    pass


class RotationAero(AeroBase):
    """
    Class for the aerotech rotation stage.
    """
    pass


class InterRotationAero(InterlockedAero, RotationAero):
    """
    Class for the interlocked aerotech rotation stage.
    """
    pass


class DiodeAero(AeroBase):
    """
    VT50 Micronix Motor of the diodes
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Exceptions for the SnD system.
"""

import logging

logger = logging.getLogger(__name__)

# Exceptions


class InputError(Exception):
    """
    Exception when the inputs to a function or method are invalid.
    """
    pass


class UndefinedBounds(Exception):
    """
    Exception when bounds of plan are not defined
    """
    pass


class SndException(Exception):
    """
    Base aerotech motor exceptions.
    """
    pass


class MotorDisabled(SndException):
    """
    Exception raised when an action requiring the motor be enabled is requested.
    """
    pass


class MotorFaulted(SndException):
    """
    Exception raised when an action requiring the motor not be faulted is
    requested.
    """
    pass


class MotorError(SndException):
    """
    Exception raised when an action requiring the motor not have an error is
    requested.
    """
    pass


class MotorStopped(SndException):
    """
    Exception raised when an action requiring the motor to not be stopped is
    requested.
    """
    pass


class BadN2Pressure(SndException):
    """
    Exception raised when an action requiring the N2 pressure be good is
    requested with a bad pressure.
    """
    pass



import logging

import numpy as np
import pytest
from bluesky.preprocessors import run_wrapper
from ophyd.sim import SynAxis

from ..plans import calibration as calib
from .conftest import SynCamera, test_df_scan

logger = logging.getLogger(__name__)

rtol = 1e-6                             # Numpy relative tolerance
m1 = SynAxis(name="m1")
m2 = SynAxis(name="m2")
delay = SynAxis(name="delay")


def test_calibration_centroid_scan_df_is_valid(fresh_RE):
    camera = SynCamera(m1, m2, delay, name="camera")

    def test_plan():
        df = yield from calib.calibration_centroid_scan(
            camera, delay, [m1, m2], -1, 1, 5, detector_fields=[
                'camera_centroid_x',
                'camera_centroid_y'])
        assert True not in df.isnull().values
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


def test_calibration_centroid_scan_raises_ValueError_correctly(fresh_RE):
    camera = SynCamera(m1, m2, delay, name="camera")

    def test_plan():
        with pytest.raises(ValueError):
            _ = yield from calib.calibration_centroid_scan(
                camera, delay, [m1, m2], -1, 1, 3, calib_fields=[m1.name],
                detector_fields=['camera_centroid_x', 'camera_centroid_y']
            )
        with pytest.raises(ValueError):
            _ = yield from calib.calibration_centroid_scan(
                camera, delay, [m1, m2], -1, 1, 3, calib_fields=[m1.name]*3,
                detector_fields=['camera_centroid_x', 'camera_centroid_y']
            )


def test_calibration_centroid_scan_renames_columns_correctly(fresh_RE):
    camera = SynCamera(m1, m2, delay, name="camera")
    detector_fields = [
        'camera_centroid_x',
        'camera_centroid_y'
    ]

    def test_plan():
        calib_motors = [m1, m2]
        df = yield from calib.calibration_centroid_scan(
            camera, delay, calib_motors, -1, 1, 3,
            detector_fields=detector_fields
        )
        expected_columns = (
            [delay.name] + [m.name + "_pre" for m in calib_motors] + detector_fields
        )
        assert (df.columns == expected_columns).all()

    fresh_RE(run_wrapper(test_plan()))


def test_detector_scaling_walk_start_positions_are_valid(fresh_RE):
    camera = SynCamera(m1, m2, delay, name="camera")
    calib_motors = [m1, m2]

    def test_plan():
        # Get the current positions of the calib motors
        expected_positions = [test_df_scan[m.name+"_pre"][-1]
                              for m in calib_motors]
        # Perform the walk
        _, start = yield from calib.detector_scaling_walk(
            test_df_scan, camera, calib_motors, tolerance=0, system=delay)

        # Make sure we dont have any bad values
        assert np.nan not in start and np.inf not in start
        # Make sure we are get the expected positions of the motors
        assert np.isclose(start, expected_positions, rtol=rtol).all()
    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


@pytest.mark.parametrize("weights", [(1, 1), (.5, -.5), (-10, 5.5)])
def test_detector_scaling_walk_scale_values_are_valid(fresh_RE, weights):
    camera = SynCamera(m1, m2, delay, name="camera")
    centroids = [camera.centroid_x, camera.centroid_y]
    calib_motors = [m1, m2]
    for cent, weight in zip(centroids, weights):
        cent.weights = [weight, cent.weights[1]]

    def test_plan():
        df_scan = yield from calib.calibration_centroid_scan(
            camera, delay, [m1, m2], -1, 1, 5, detector_fields=[
                'camera_centroid_x',
                'camera_centroid_y'])

        # Get the expected scales if we do the walk
        expected_scales = [1/cent.weights[0] for cent in centroids]
        # Perform the walk
        scales, _ = yield from calib.detector_scaling_walk(
            df_scan, camera, calib_motors, tolerance=0, system=delay)

        # Make sure we dont have any bad values
        assert np.nan not in scales and np.inf not in scales
        # Make sure we are get the expected positions of the motors
        assert np.isclose(scales, expected_scales, rtol=rtol).all()

    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


def test_build_calibration_df_creates_correct_df_columns(fresh_RE):
    test_scale = [1.0, 1.0]
    test_start = [0.25, -0.25]
    camera = SynCamera(m1, m2, delay, name="camera")
    # The resulting df should be the same columns with two new ones
    new_columns = [m.name+"_post" for m in [m1, m2]]

    scan_columns = list(test_df_scan.columns)
    expected_columns = [delay.name] + new_columns

    df_calib = calib.build_calibration_df(test_df_scan, test_scale, test_start,
                                          camera)
    # Make sure we didn't mutate the original scan df
    assert (test_df_scan.columns == scan_columns).all()
    # Make sure we get what we expect
    assert (df_calib.columns == expected_columns).all()


@pytest.mark.parametrize("weights", [(1, 1), (.5, -.5), (-10, 5.5)])
def test_scale_scan_df_creates_correct_calibration_tables(fresh_RE, weights):
    camera = SynCamera(m1, m2, delay, name="camera")
    centroids = [camera.centroid_x, camera.centroid_y]
    calib_motors = [m1, m2]
    for cent, weight in zip(centroids, weights):
        cent.weights = [weight, cent.weights[1]]

    def test_plan():
        # Perform the scan
        df_scan = yield from calib.calibration_centroid_scan(
            camera, delay, [m1, m2], -1, 1, 5, detector_fields=[
                'camera_centroid_x',
                'camera_centroid_y'])

        # Perform the walk
        scales, starts = yield from calib.detector_scaling_walk(
            df_scan, camera, calib_motors, tolerance=0, system=delay)

        # Get the scaled scan dataframe
        df_calib = calib.build_calibration_df(df_scan, scales, starts, camera)

        # Expected positions of the centroids are the first positions
        expected_centroids = df_scan[[c.name for c in centroids]].iloc[0]

        for i in range(len(df_scan)):
            # Move to the scan position for the main motor
            delay.set(df_calib[delay.name].iloc[i])

            for cmotor, exp_cent, cent in zip(calib_motors, expected_centroids,
                                              centroids):
                # Move to the abosolute corrected position
                cmotor.set(df_calib[cmotor.name+"_post"].iloc[i])
                # Check the centroids are where they should be
                assert np.isclose(cent.get(), exp_cent, rtol=rtol)

    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


@pytest.mark.parametrize("weights", [(1, 1), (.5, -.5), (-10, 5.5)])
def test_calibration_scan(fresh_RE, weights):
    camera = SynCamera(m1, m2, delay, name="camera")
    centroids = [camera.centroid_x, camera.centroid_y]
    calib_motors = [m1, m2]
    for cent, weight in zip(centroids, weights):
        cent.weights = [weight, cent.weights[1]]

    def test_plan():
        # Perform the scan
        df_calib, df_scan, _, _ = yield from calib.calibration_scan(
            camera, ['camera_centroid_x', 'camera_centroid_y'], delay, None,
            [m1, m2], None, -1, 1, 5, tolerance=0)

        # Expected positions of the centroids are the first positions
        expected_centroids = df_scan[[c.name for c in centroids]].iloc[0]

        for i in range(len(df_scan)):
            # Move to the scan position for the main motor
            delay.set(df_calib[delay.name].iloc[i])

            for cmotor, exp_cent, cent in zip(calib_motors, expected_centroids,
                                              centroids):
                # Move to the abosolute corrected position
                cmotor.set(df_calib[cmotor.name+"_post"].iloc[i])
                # Check the centroids are where they should be
                assert np.isclose(cent.get(), exp_cent, rtol=rtol)

    # Run the plan
    fresh_RE(run_wrapper(test_plan()))


def test_calibrate_motor(fresh_RE, get_calib_motor):
    # Should be properly configured to start
    motor = get_calib_motor
    camera = SynCamera(*motor.calib_motors, motor, name="camera")
    calib_motors = motor.calib_motors
    start, stop, steps = -1, 1, 5

    def test_plan():
        # Perform the scan to get the expected scan results
        df_calib, df_scan, scale, start_pos = yield from calib.calibration_scan(
            camera,
            motor.detector_fields,
            motor,
            motor.motor_fields,
            motor.calib_motors,
            motor.calib_fields,
            start, stop, steps,
            tolerance=0)

        # Calibrate the motor
        _, _ = yield from calib.calibrate_motor(
            camera,
            motor.detector_fields,
            motor,
            motor.motor_fields,
            motor.calib_motors,
            motor.calib_fields,
            start, stop, steps,
            confirm_overwrite=False,
            tolerance=0)

        # Get the configuration from the motor as it sees it
        config = motor.read_configuration()

        # Ensure the resulting config is what we got in the initial scan
        assert config['calib']['value'].equals(df_calib)
        assert config['scan']['value'].equals(df_scan)
        assert config['motors']['value'] == [motor] + calib_motors
        assert config['scale']['value'] == scale
        assert config['start']['value'] == start_pos

    # Run the plan
    fresh_RE(run_wrapper(test_plan()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from collections import OrderedDict

import pytest
from ophyd.device import Device

from hxrsnd import rtd

from .conftest import fake_device, get_classes_in_module

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("dev", get_classes_in_module(rtd, Device))
def test_rtd_devices_instantiate_and_run_ophyd_functions(dev):
    device = fake_device(dev)
    assert(isinstance(device.read(), OrderedDict))
    assert(isinstance(device.read_configuration(), OrderedDict))



import logging

import pytest
from bluesky.plan_stubs import rel_set
from bluesky.preprocessors import run_wrapper
from ophyd.sim import SynAxis

from ..plans.preprocessors import return_to_start

logger = logging.getLogger(__name__)

m1 = SynAxis(name="m1")
m2 = SynAxis(name="m2")
motors = [m1, m2]


@pytest.mark.parametrize("return_devices", [[], [m1], [m1, m2]])
def test_return_to_start_returns_motors_correctly(fresh_RE, return_devices):
    # Grab the initial positions
    initial_positions = [mot.position for mot in motors]
    final_positions = []
    # Create the plan

    @return_to_start(*return_devices)
    def test_plan():
        for mot in motors:
            yield from rel_set(mot, 1)
            # Grab the final positions
            final_positions.append(mot.position)

    # Run the plan
    fresh_RE(run_wrapper(test_plan()))

    # get the expected and current positions
    expected_positions = [ipos if m in return_devices else fpos
                          for m, ipos, fpos in zip(
                              motors, initial_positions, final_positions)]
    current_positions = [mot.position for mot in motors]

    # Assert they are the same
    assert current_positions == expected_positions



from ._version import get_versions
from .plans.alignment import maximize_lorentz, rocking_curve

__version__ = get_versions()['version']
del get_versions

__all__ = ['maximize_lorentz', 'rocking_curve']



from setuptools import find_packages, setup

import versioneer

setup(name='hxrsnd',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license='BSD',
      author='SLAC National Accelerator Laboratory',
      packages=find_packages(),
      include_package_data=True,
      description='Python controls suite for HXRSnD',
      )



"""
Common SnD device classes
"""
import logging

from ophyd.device import Device
from pcdsdevices.interface import BaseInterface

logger = logging.getLogger(__name__)


class SndDevice(BaseInterface, Device):
    """
    Base Sndmotor class
    """

    tab_component_names = True
    tab_whitelist = ['st']

    def __init__(self, prefix, name=None, desc=None, set_timeout=1, *args,
                 **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.desc = desc or self.name
        self.set_timeout = set_timeout

    def _apply_all(self, method, subclass=object, *method_args,
                   **method_kwargs):
        """
        Runs the method for all devices that are of the inputted subclass. All
        additional arguments and key word arguments are passed as inputs to the
        method.

        Parameters
        ----------
        method : str
            Method of each device to run.

        subclass : class
            Subclass to run the methods for.

        method_args : tuple, optional
            Positional arguments to pass to the method

        method_kwargs : dict, optional
            Key word arguments to pass to the method
        """
        ret = []
        # Check if each component is a subclass of subclass then run the method
        for comp_name in self.component_names:
            component = getattr(self, comp_name)
            if issubclass(type(component), subclass):
                ret.append(getattr(component, method)(*method_args,
                                                      **method_kwargs))
        return ret

    def st(self, *args, **kwargs):
        """
        Returns or prints the status of the device. Alias for 'device.status()'.

        Parameters
        ----------
        print_status : bool, optional
            Determines whether the dataframe should be printed or returned

        short : bool, optional
            Use a shortened list or all relevant parameters
        """
        return self.status(*args, **kwargs)

    def __repr__(self):
        """
        Returns the status of the device. Alias for status().

        Returns
        -------
        status : str
            Status string.
        """
        # Try to return the status
        try:
            return self.status(print_status=False)
        # There is no scenario where we would want to know of an error here
        except Exception:
            return super().__repr__()



"""
Tests for pyutils.pyutils
"""
import logging
from collections.abc import Iterable
from math import isnan
from pathlib import Path

import numpy as np
import pytest

from hxrsnd import utils

logger = logging.getLogger(__name__)

test_values = [2, np.pi, True, "test_s", "10", ["test"], ("test",), {"test": 1}]
test_lists = [[1, 2, 3, 4, 5], [[1], [2], [3], [4], [5]], [[1, 2, 3], [4, 5]],
              [[1, [2, [3, [4, [5]]]]]]]


@pytest.mark.parametrize("test", test_values)
def test_isiterable_correctly_returns(test):
    iterable = utils.isiterable(test)
    if isinstance(test, str):
        assert iterable is False
    elif isinstance(test, Iterable):
        assert iterable is True
    else:
        assert iterable is False


@pytest.mark.parametrize("test", test_lists)
def test_flatten_works_correctly(test):
    assert utils.flatten(test) == [1, 2, 3, 4, 5]


@pytest.mark.parametrize("screen", ["motor_expert_screen.sh", "snd_main"])
def test_absolute_submodule_path_works_correctly(screen):
    path = "HXRSnD/screens/{0}".format(screen)
    template = "/reg/neh/operator/xcsopr/bin/snd/HXRSnD/hxrsnd/utils.py"
    abs_path = Path(utils.absolute_submodule_path(path, template))
    assert abs_path == Path("/".join(template.split("/")[:-3]) + "/" + path)


def test_stop_on_keyboardinterrupt_runs_stop_method():
    class TestClass:
        name = "test"
        stopped = False

        @utils.stop_on_keyboardinterrupt
        def something_that_raises_keyboardinterrupt(self):
            raise KeyboardInterrupt

        def stop(self):
            self.stopped = True
    tst = TestClass()
    assert tst.stopped is False
    tst.something_that_raises_keyboardinterrupt()
    assert tst.stopped is True


def test_nan_if_no_parent_works_for_methods_and_properties():
    class Test:
        @property
        @utils.nan_if_no_parent
        def tst_property(self):
            return True

        @utils.nan_if_no_parent
        def tst_method(self):
            return True
    tst = Test()
    tst.parent = None
    assert isnan(tst.tst_property)
    assert isnan(tst.tst_method())
    tst.parent = True
    assert tst.tst_property is True
    assert tst.tst_method() is True



"""
Diodes
"""
import logging

import numpy as np
from ophyd import EpicsSignalRO
from ophyd.device import Component as C
from ophyd.device import FormattedComponent as FC

from .aerotech import DiodeAero
from .snddevice import SndDevice

logger = logging.getLogger(__name__)


class DiodeBase(SndDevice):
    """
    Base class for the diode.
    """
    pass


class HamamatsuDiode(DiodeBase):
    """
    Class for the Hamamatsu diode.
    """
    pass


class HamamatsuXMotionDiode(SndDevice):
    """
    Class for the Hamamatsu diode but with an X motor
    """

    tab_component_names = True
    tab_whitelist = ['block', 'blocked', 'unblock']

    diode = C(HamamatsuDiode, ":DIODE")
    x = C(DiodeAero, ":X")

    def __init__(self, prefix, name=None, block_pos=5, unblock_pos=0, *args,
                 block_atol=0.001, desc=None, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.block_pos = block_pos
        self.unblock_pos = unblock_pos
        self.block_atol = block_atol
        self.desc = desc or self.name

    @property
    def blocked(self):
        """
        Returns if the diode is in the blocked position.
        """
        if np.isclose(self.x.position, self.block_pos, atol=self.block_atol):
            return True
        elif np.isclose(self.x.position, self.unblock_pos,
                        atol=self.block_atol):
            return False
        else:
            return "Unknown"

    def block(self, *args, **kwargs):
        """
        Moves the diode into the blocking position.

        Parameters
        ----------
        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        ret_status : bool, optional
            Return the status object of the move.

        print_move : bool, optional
            Print a short statement about the move.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        return self.x.mv(self.block_pos, *args, **kwargs)

    def unblock(self, *args, **kwargs):
        """
        Moves the diode into the nonblocking position.

        Parameters
        ----------
        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        ret_status : bool, optional
            Return the status object of the move.

        print_move : bool, optional
            Print a short statement about the move.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        return self.x.mv(self.unblock_pos, *args, **kwargs)


class HamamatsuXYMotionCamDiode(SndDevice):
    """
    Class for the Hamamatsu diode but with X and Y motors
    """
    tab_component_names = True
    tab_whitelist = ['block', 'blocked', 'unblock']

    diode = C(HamamatsuDiode, ":DIODE")
    x = C(DiodeAero, ":X")
    y = C(DiodeAero, ":Y")

    def __init__(self, prefix, name=None, block_pos=5, pos_func=None,
                 block_atol=0.001, desc=None, *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.block_pos = block_pos
        self.pos_func = pos_func
        self.block_atol = block_atol
        self.desc = desc or self.name

    @property
    def blocked(self):
        """
        Returns if the diode is in the blocked position.

        Returns
        -------
        blocked : bool or str
            True or False if it is close to the blocked or unblocked positions.
            Returns 'Unknown' if it is far from either of those positions.
        """
        if callable(self.pos_func):
            if np.isclose(self.x.position, self.pos_func()+self.block_pos,
                          atol=self.block_atol):
                return True
            elif np.isclose(self.x.position, self.pos_func(),
                            atol=self.block_atol):
                return False
        return "Unknown"

    def block(self, *args, **kwargs):
        """
        Moves the diode by the blocking position defined by the position
        function plus the block position.

        Parameters
        ----------
        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        ret_status : bool, optional
            Return the status object of the move.

        print_move : bool, optional
            Print a short statement about the move.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        # Move to the blocked position if we aren't already there
        if self.blocked is True:
            # We are already in the blocked position
            logger.info("Motor '{0}' is currently in the blocked position"
                        "".format(self.x.desc))
        else:
            return self.x.mv(self.pos_func() + self.block_pos, *args, **kwargs)

    def unblock(self, *args, **kwargs):
        """
        Moves the diode by the nonblocking position defined by the position
        function

        Parameters
        ----------
        wait : bool, optional
            Wait for the motor to complete the motion.

        check_status : bool, optional
            Check if the motors are in a valid state to move.

        ret_status : bool, optional
            Return the status object of the move.

        print_move : bool, optional
            Print a short statement about the move.

        Returns
        -------
        status : MoveStatus
            Status object for the move.
        """
        # Move to the blocked position if we aren't already there
        if self.blocked is False:
            # We are already in the blocked position
            logger.info("Motor '{0}' is currently in the unblocked position"
                        "".format(self.x.desc))
        else:
            return self.x.mv(self.pos_func(), *args, **kwargs)


class DiodeIO(SndDevice):
    """
    Peak information for a Wave8 input

    Parameters
    ----------
    prefix : str
        Base name of device

    channel : int
        Channel number of device

    name : str
        Name of Wave8 device
    """
    tab_component_names = True

    peakA = FC(EpicsSignalRO, '{self.prefix}:_peakA_{self.channel}')
    peakT = FC(EpicsSignalRO, '{self.prefix}:_peakT_{self.channel}')

    def __init__(self, prefix, channel, name, *,
                 read_attrs=None, **kwargs):
        # Store the channel
        self.channel = channel
        # Default read attributes
        if read_attrs is None:
            read_attrs = ['peakT']
        # Initialize device
        super().__init__(prefix, name=name, read_attrs=read_attrs, **kwargs)


class Wave8(SndDevice):
    """
    Wave8 Device

    A system of sixteen diodes, each with two peaks; A and T.
    """
    tab_component_names = True
    diode_0 = C(DiodeIO, '', channel=0, name='Diode 0')
    diode_1 = C(DiodeIO, '', channel=1, name='Diode 1')
    diode_2 = C(DiodeIO, '', channel=2, name='Diode 2')
    diode_3 = C(DiodeIO, '', channel=3, name='Diode 3')
    diode_4 = C(DiodeIO, '', channel=4, name='Diode 4')
    diode_5 = C(DiodeIO, '', channel=5, name='Diode 5')
    diode_6 = C(DiodeIO, '', channel=6, name='Diode 6')
    diode_7 = C(DiodeIO, '', channel=7, name='Diode 7')
    diode_8 = C(DiodeIO, '', channel=8, name='Diode 8')
    diode_9 = C(DiodeIO, '', channel=9, name='Diode 9')
    diode_10 = C(DiodeIO, '', channel=10, name='Diode 10')
    diode_11 = C(DiodeIO, '', channel=11, name='Diode 11')
    diode_12 = C(DiodeIO, '', channel=12, name='Diode 12')
    diode_13 = C(DiodeIO, '', channel=13, name='Diode 13')
    diode_14 = C(DiodeIO, '', channel=14, name='Diode 14')
    diode_15 = C(DiodeIO, '', channel=15, name='Diode 15')



"""
Script to hold the split and delay class.

All units of time are in picoseconds, units of length are in mm.
"""
import logging

from ophyd import Component as Cmp

from .diode import HamamatsuXMotionDiode, HamamatsuXYMotionCamDiode
from .macromotor import DelayMacro, Energy1CCMacro, Energy1Macro, Energy2Macro
from .pneumatic import SndPneumatics
from .snddevice import SndDevice
from .tower import ChannelCutTower, DelayTower
from .utils import absolute_submodule_path

logger = logging.getLogger(__name__)


class SplitAndDelay(SndDevice):
    """
    Hard X-Ray Split and Delay System.

    Components

    t1 : DelayTower
        Tower 1 in the split and delay system.

    t4 : DelayTower
        Tower 4 in the split and delay system.

    t2 : ChannelCutTower
        Tower 2 in the split and delay system.

    t3 : ChannelCutTower
        Tower 3 in the split and delay system.

    ab : SndPneumatics
        Vacuum device object for the system.

    di : HamamatsuXYMotionCamDiode
        Input diode for the system.

    dd : HamamatsuXYMotionCamDiode
        Diode between the two delay towers.

    do : HamamatsuXYMotionCamDiode
        Output diode for the system.

    dci : HamamatsuXMotionDiode
        Input diode for the channel cut line.

    dcc : HamamatsuXMotionDiode
        Diode between the two channel cut towers.

    dco : HamamatsuXMotionDiode
        Input diode for the channel cut line.

    E1 : Energy1Macro
        Delay energy pseudomotor.

    E2 : Energy2Macro
        Channel cut energy pseudomotor.

    delay : DelayMacro
        Delay pseudomotor.
    """
    tab_component_names = True
    tab_whitelist = ['st', 'status', 'diag_status', 'theta1', 'theta2',
                     'main_screen', 'status']
    # Delay Towers
    t1 = Cmp(DelayTower, ":T1", pos_inserted=21.1, pos_removed=0,
             desc="Tower 1")
    t4 = Cmp(DelayTower, ":T4", pos_inserted=21.1, pos_removed=0,
             desc="Tower 4")

    # Channel Cut Towers
    t2 = Cmp(ChannelCutTower, ":T2", pos_inserted=None, pos_removed=0,
             desc="Tower 2")
    t3 = Cmp(ChannelCutTower, ":T3", pos_inserted=None, pos_removed=0,
             desc="Tower 3")

    # Pneumatic Air Bearings
    ab = Cmp(SndPneumatics, "")

    # SnD and Delay line diagnostics
    di = Cmp(HamamatsuXMotionDiode, ":DIA:DI", desc="DI")
    dd = Cmp(HamamatsuXYMotionCamDiode, ":DIA:DD", desc="DD")
    do = Cmp(HamamatsuXMotionDiode, ":DIA:DO", desc="DO")

    # Channel Cut Diagnostics
    dci = Cmp(HamamatsuXMotionDiode, ":DIA:DCI", block_pos=-5, desc="DCI")
    dcc = Cmp(HamamatsuXYMotionCamDiode, ":DIA:DCC", block_pos=-5, desc="DCC")
    dco = Cmp(HamamatsuXMotionDiode, ":DIA:DCO", block_pos=-5, desc="DCO")

    # Macro motors
    E1 = Cmp(Energy1Macro, "", desc="Delay Energy")
    E1_cc = Cmp(Energy1CCMacro, "", desc="CC Delay Energy")
    E2 = Cmp(Energy2Macro, "", desc="CC Energy")
    delay = Cmp(DelayMacro, "", desc="Delay")

    def __init__(self, prefix, name=None, daq=None, RE=None, *args, **kwargs):
        super().__init__(prefix, name=name, *args, **kwargs)
        self.daq = daq
        self.RE = RE
        self._delay_towers = [self.t1, self.t4]
        self._channelcut_towers = [self.t2, self.t3]
        self._towers = self._delay_towers + self._channelcut_towers
        self._delay_diagnostics = [self.di, self.dd, self.do]
        self._channelcut_diagnostics = [self.dci, self.dcc, self.dco]
        self._diagnostics = self._delay_diagnostics+self._channelcut_diagnostics

        # Set the position calculators of dd and dcc
        self.dd.pos_func = lambda: \
            self.E1._get_delay_diagnostic_position()
        self.dcc.pos_func = lambda: \
            self.E2._get_channelcut_diagnostic_position()

    def diag_status(self):
        """
        Prints a string containing the blocking status and the position of the
        motor.
        """
        status = "\n{0}{1:<14}|{2:^16}|{3:^16}\n{4}{5}".format(
            " "*2, "Diagnostic", "Blocking", "Position", " "*2, "-"*50)
        for diag in self._diagnostics:
            status += "\n{0}{1:<14}|{2:^16}|{3:^16.3f}".format(
                " "*2, diag.desc, str(diag.blocked), diag.x.position)
        logger.info(status)

    @property
    def theta1(self):
        """
        Returns the bragg angle the the delay line is currently set to
        maximize.

        Returns
        -------
        theta1 : float
            The bragg angle the delay line is currently set to maximize
            in degrees.
        """
        return self.t1.theta

    @property
    def theta2(self):
        """
        Returns the bragg angle the the delay line is currently set to
        maximize.

        Returns
        -------
        theta2 : float
            The bragg angle the channel cut line is currently set to maximize
            in degrees.
        """
        return self.t2.theta

    def main_screen(self, print_msg=True):
        """
        Launches the main SnD screen.
        """
        # Get the absolute path to the screen
        path = absolute_submodule_path("hxrsnd/screens/snd_main")
        if print_msg:
            logger.info("Launching expert screen.")
        logger.error('TODO - not yet implemented (path: %s)', path)
        # NOTE: this was the command ,where `p` and `axis` were not defined:
        # os.system("{0} {1} {2} &".format(path, p, axis))

    def status(self, print_status=True):
        """
        Returns the status of the split and delay system.

        Returns
        -------
        Status : str
        """
        status = "Split and Delay System Status\n"
        status += "-----------------------------"
        status = self.E1.status(status, 0, print_status=False)
        status = self.E2.status(status, 0, print_status=False)
        status = self.delay.status(status, 0, print_status=False, newline=True)
        status = self.t1.status(status, 0, print_status=False, newline=True)
        status = self.t2.status(status, 0, print_status=False, newline=True)
        status = self.t3.status(status, 0, print_status=False, newline=True)
        status = self.t4.status(status, 0, print_status=False, newline=True)
        status = self.ab.status(status, 0, print_status=False, newline=False)

        if print_status:
            logger.info(status)
        else:
            logger.debug(status)
            return status
